# Copyright Cartopy Contributors
#
# This file is part of Cartopy and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tests for specific Cartopy CRS subclasses.

"""

import pytest


pytest.register_assert_rewrite('cartopy.tests.crs.helpers')
